
# THIS FILE IS GENERATED FROM ORANGE3-TEXT SETUP.PY
short_version = '0.7.4'
version = '0.7.4'
full_version = '0.7.4'
git_revision = 'ed1a1317473c568ad6296608840a1be96e0e72b6'
release = True
if not release:
    version = full_version
    short_version += ".dev"
